"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.YouTube = void 0;
const node_fetch_1 = require("node-fetch");
const Channel_1 = require("./structures/Channel");
const Playlist_1 = require("./structures/Playlist");
const Video_1 = require("./structures/Video");
const Resource_1 = require("./types/Resource");
class YouTube {
    constructor(key, options) {
        this.key = key;
        this.options = {
            cache: (options === null || options === void 0 ? void 0 : options.cache) || true,
            fetchAll: (options === null || options === void 0 ? void 0 : options.fetchAll) || false
        };
    }
    async getResource(type, qs = {}) {
        qs = { part: Resource_1.ResourcePart[type], ...qs };
        const result = await this.request(Resource_1.ResourceEndpoint[type], qs);
        if (result.items.length > 0)
            return result.items[0];
        throw new Error(`Resource ${result.kind || type} not found`);
    }
    async getResourceByID(type, id, qs = {}) {
        return this.getResource(type, Object.assign(qs, { id }));
    }
    async getChannel(url, options = {}) {
        const id = Channel_1.Channel.extractId(url);
        if (!id)
            throw new Error(`No channel ID found in URL: ${url}`);
        return this.getChannelById(id, options);
    }
    async getChannelById(id, options = {}) {
        const result = await this.getResourceByID('channel', id, options);
        if (result)
            return new Channel_1.Channel(this, result);
        return null;
    }
    async getPlaylist(url, options = {}) {
        const id = Playlist_1.Playlist.extractId(url);
        if (!id)
            throw new Error(`No playlist ID found in URL: ${url}`);
        return this.getPlaylistById(id, options);
    }
    async getPlaylistById(id, options = {}) {
        const result = await this.getResourceByID('playlist', id, options);
        if (result)
            return new Playlist_1.Playlist(this, result);
        return null;
    }
    async getVideo(url, options = {}) {
        const id = Video_1.Video.extractId(url);
        if (!id)
            throw new Error(`No video ID found in URL: ${url}`);
        return this.getVideoById(id, options);
    }
    async getVideoById(id, options = {}) {
        const result = await this.getResourceByID('video', id, options);
        if (result)
            return new Video_1.Video(this, result);
        return null;
    }
    async search(query, limit = 5, options = {}) {
        const result = await this.paginatedRequest(Resource_1.ResourceEndpoint.search, limit, Object.assign(options, { q: query, part: Resource_1.ResourcePart.search }));
        return result.map((item) => {
            if (item.id.kind === Resource_1.ResourceKind.video)
                return new Video_1.Video(this, item);
            if (item.id.kind === Resource_1.ResourceKind.playlist)
                return new Playlist_1.Playlist(this, item);
            if (item.id.kind === Resource_1.ResourceKind.channel)
                return new Channel_1.Channel(this, item);
            return undefined;
        });
    }
    async searchChannels(query, limit = 5, options = {}) {
        return this.search(query, limit, Object.assign(options, { type: 'channel' }));
    }
    async searchPlaylists(query, limit = 5, options = {}) {
        return this.search(query, limit, Object.assign(options, { type: 'playlist' }));
    }
    async searchVideos(query, limit = 5, options = {}) {
        return this.search(query, limit, Object.assign(options, { type: 'video' }));
    }
    async request(endpoint, qs = {}) {
        qs = { key: this.key, ...qs };
        const parameters = Object.keys(qs).filter((k) => qs[k]).map((k) => `${k}=${qs[k]}`);
        const res = await node_fetch_1.default(encodeURI(`https://www.googleapis.com/youtube/v3/${endpoint}${parameters.length > 0 ? `?${parameters.join('&')}` : ''}`));
        const result = await res.json();
        if (result.error) {
            if (result.error.code === 400) {
                if (result.error.errors && result.error.errors.length > 0) {
                    throw new Error(`YouTube API Error: ${result.error.errors[0].domain} - ${result.error.errors[0].reason}`);
                }
            }
            else
                throw new Error(`An unknown YouTube API error occurred: ${result.error.code || result.error}`);
        }
        return result;
    }
    async paginatedRequest(endpoint, count = 50, options = {}, fetched = [], pageToken) {
        if (count < 1)
            throw new Error('Cannot fetch less than 1.');
        const limit = count > 50 ? 50 : count;
        const res = await this.request(endpoint, { maxResults: limit, pageToken, ...options });
        const results = fetched.concat(res.items);
        if (res.nextPageToken && limit !== count)
            return this.paginatedRequest(endpoint, count - limit, options, results, res.nextPageToken);
        return results;
    }
}
exports.YouTube = YouTube;
